from dataclasses import dataclass

from .resource import Resource, ResourceSpec, ResourceNameTrait
from .role import Role
from ..enums import AccountEdition, ResourceType
from ..scope import SchemaScope
from ..props import Props, StringProp, StringListProp
from ..resource_name import ResourceName


@dataclass(unsafe_hash=True)
class _Tag(ResourceSpec):
    name: ResourceName
    comment: str = None
    allowed_values: list = None


class Tag(ResourceNameTrait, Resource):
    """
    Description:
        Represents a tag in Snowflake, which can be used to label various resources for better management and categorization.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-tag

    Fields:
        name (string, required): The name of the tag.
        allowed_values (list): A list of allowed values for the tag.
        comment (string): A comment or description for the tag.

    Python:

        ```python
        tag = Tag(
            name="cost_center",
            allowed_values=["finance", "engineering", "sales"],
            comment="This is a sample tag",
        )
        ```

    Yaml:

        ```yaml
        tags:
          - name: cost_center
            comment: This is a sample tag
            allowed_values:
              - finance
              - engineering
              - sales
        ```
    """

    edition = {AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    resource_type = ResourceType.TAG
    props = Props(
        allowed_values=StringListProp("allowed_values", eq=False, parens=False),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _Tag

    def __init__(
        self,
        name: str,
        comment: str = None,
        allowed_values: list = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _Tag = _Tag(
            name=self._name,
            comment=comment,
            allowed_values=allowed_values,
        )



from typing import Optional, Union

from .enums import ResourceType

from .resource_name import ResourceName


def _params_to_str(params: dict) -> str:
    return "&".join([f"{k.lower()}={v}" for k, v in params.items()])


def resource_label_for_type(resource_type: ResourceType) -> str:
    return str(resource_type).replace(" ", "_").lower()


def resource_type_for_label(resource_label: str) -> ResourceType:
    return ResourceType(resource_label.upper().replace("_", " "))


def names_are_equal(name1: Union[None, str, ResourceName], name2: Union[None, str, ResourceName]) -> bool:
    if name1 is None and name2 is None:
        return True
    if name1 is None or name2 is None:
        return False
    return ResourceName(name1) == ResourceName(name2)


class FQN:
    def __init__(
        self,
        name: str,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        arg_types: Optional[list] = None,
        params: Optional[dict] = None,
    ) -> None:
        self.name = name
        self.database = database
        self.schema = schema
        self.arg_types = arg_types or []
        self.params = params or {}

    def __eq__(self, other):
        if not isinstance(other, FQN):
            return False
        return (
            names_are_equal(self.name, other.name)
            and names_are_equal(self.database, other.database)
            and names_are_equal(self.schema, other.schema)
            and self.arg_types == other.arg_types
            and self.params == other.params
        )

    def __hash__(self):
        return hash(
            (
                ResourceName(self.name),
                ResourceName(self.database) if self.database else None,
                ResourceName(self.schema) if self.schema else None,
                tuple(self.arg_types),
                tuple(
                    self.params.items(),
                ),
            )
        )

    def __str__(self):
        db = f"{ResourceName(self.database)}." if self.database else ""
        schema = f"{ResourceName(self.schema)}." if self.schema else ""
        arg_types = f"({', '.join(self.arg_types)})" if self.arg_types else ""
        params = "?" + _params_to_str(self.params) if self.params else ""
        return f"{db}{schema}{self.name}{arg_types}{params}"

    def __repr__(self):  # pragma: no cover
        db = f", db={self.database}" if self.database else ""
        schema = f", schema={self.schema}" if self.schema else ""
        params = "?" + _params_to_str(self.params) if self.params else ""
        return f"FQN(name={self.name}{db}{schema}{params})"


class URN:
    """
    Universal Resource Name

    An address scheme for uniquely identifying resources within a Snowflake account.

    Format
    ------

                     Resource
              Account  Type         Resource
          Org     │     │            Name     Params
        ───┴── ───┴── ──┴──        ───┴───── ───┴───────
    urn:ABC123:XYZ987:table/db.sch.sometable?param=value
                            ───┬────────────
                             Fully Qualified Name
    """

    def __init__(self, resource_type: ResourceType, fqn: FQN, account_locator: str) -> None:
        if not isinstance(resource_type, ResourceType):
            raise Exception(f"Invalid resource type: {resource_type}")
        self.resource_type: ResourceType = resource_type
        self.resource_label: str = resource_label_for_type(resource_type)
        self.fqn: FQN = fqn
        self.account_locator: str = account_locator
        self.organization: str = ""

    def __eq__(self, other):
        if not isinstance(other, URN):
            return False
        return (
            self.resource_type == other.resource_type
            and self.fqn == other.fqn
            and self.account_locator == other.account_locator
        )

    def __hash__(self):
        return hash((self.resource_type, self.fqn, self.account_locator))

    def __str__(self):
        return f"urn:{self.organization}:{self.account_locator}:{self.resource_label}/{self.fqn}"

    def __repr__(self):  # pragma: no cover
        org = getattr(self, "organization", "")
        acct = getattr(self, "account_locator", "")
        label = getattr(self, "resource_label", "")
        fqn = getattr(self, "fqn", "")
        return f"URN(urn:{org}:{acct}:{label}/{fqn})"

    @classmethod
    def from_resource(cls, resource, account_locator: str = ""):
        return cls(resource_type=resource.resource_type, fqn=resource.fqn, account_locator=account_locator)

    # @classmethod
    # def from_locator(cls, locator: "ResourceLocator"):
    #     if locator.star:
    #         raise Exception("Cannot create URN from a wildcard locator")
    #     return cls(resource_type=locator.resource_key, fqn=FQN.from_str(locator.locator))

    @classmethod
    def from_session_ctx(cls, session_ctx):
        return cls(
            resource_type=ResourceType.ACCOUNT,
            fqn=FQN(name=session_ctx["account"]),
            account_locator=session_ctx["account_locator"],
        )

    def database(self):
        if not self.fqn.database:
            raise Exception(f"URN does not have a database: {self}")
        return URN(
            resource_type=ResourceType.DATABASE,
            account_locator=self.account_locator,
            fqn=FQN(name=self.fqn.database),
        )

    def schema(self):
        if not self.fqn.schema:
            raise Exception(f"URN does not have a schema: {self}")
        return URN(
            resource_type=ResourceType.SCHEMA,
            account_locator=self.account_locator,
            fqn=FQN(name=self.fqn.schema, database=self.fqn.database),
        )


class ResourceLocator:
    """
    ResourceLocator

    A simple query language for locating resources within a Snowflake account.
    """

    def __init__(self, resource_key: str, locator: str) -> None:
        self.resource_key = resource_key
        self.locator = locator
        self.star = self.locator == "*"

    @classmethod
    def from_str(cls, resource_str: str) -> "ResourceLocator":
        """
        Parse a resource locator string.

        Usage
        -----
        Locate all resources:
        >>> ResourceLocator.from_str("*")

        Locate a specific resource:
        >>> ResourceLocator.from_str("database:mydb")
        >>> ResourceLocator.from_str("schema:mydb.my_schema")
        >>> ResourceLocator.from_str("table:mydb.my_schema.my_table")

        Locate all resources of a given type:
        >>> ResourceLocator.from_str("database:*")

        Locate all resources within a given scope:
        >>> ResourceLocator.from_str("database:mydb.*")
        """

        if resource_str == "*":
            return cls(resource_key="account", locator="*")

        parts = resource_str.split(":")
        if len(parts) != 2:
            raise Exception(f"Invalid resource locator string: {resource_str}")
        return cls(resource_key=parts[0], locator=parts[1])

    def __str__(self):
        return f"{self.resource_key}:{self.locator}"

    def __repr__(self):  # pragma: no cover
        return f"ResourceLocator(resource_key='{self.resource_key}', locator='{self.locator}')"



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, ResourceNameTrait
from .role import Role
from ..enums import BinaryFormat, Compression, FileType, ResourceType
from ..props import (
    BoolProp,
    EnumProp,
    IntProp,
    Props,
    StringListProp,
    StringProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope


@dataclass(unsafe_hash=True)
class _CSVFileFormat(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    type: FileType = FileType.CSV
    compression: Compression = None
    record_delimiter: str = "\n"
    field_delimiter: str = None
    file_extension: str = None
    parse_header: bool = False
    skip_header: int = None
    skip_blank_lines: bool = False
    date_format: str = "AUTO"
    time_format: str = "AUTO"
    timestamp_format: str = "AUTO"
    binary_format: BinaryFormat = BinaryFormat.HEX
    escape: str = None
    escape_unenclosed_field: str = "\\"
    trim_space: bool = False
    field_optionally_enclosed_by: str = None
    null_if: list[str] = None
    error_on_column_count_mismatch: bool = True
    replace_invalid_characters: bool = False
    empty_field_as_null: bool = None
    skip_byte_order_mark: bool = True
    encoding: str = "UTF8"
    comment: str = None


class CSVFileFormat(ResourceNameTrait, Resource):
    """
    Description:
        Defines the specifications for a CSV file format in Snowflake, including delimiters, encoding, and compression options.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-file-format

    Fields:
        type (FileType, required): The type of file format, which is CSV for this class. Defaults to CSV.
        compression (Compression): The compression type used for the file format.
        record_delimiter (string): Specifies the character that delimits records. Defaults to "\n".
        field_delimiter (string): Specifies the character that delimits fields.
        file_extension (string): The file extension used for files of this format.
        parse_header (bool): Whether to parse the first line of the file as a header. Defaults to False.
        skip_header (int): The number of header lines to skip before parsing.
        skip_blank_lines (bool): Whether to skip over blank lines. Defaults to False.
        date_format (string): The format used for date values. Defaults to "AUTO".
        time_format (string): The format used for time values. Defaults to "AUTO".
        timestamp_format (string): The format used for timestamp values. Defaults to "AUTO".
        binary_format (BinaryFormat): The format used for binary data. Defaults to HEX.
        escape (string): The escape character used in the file format.
        escape_unenclosed_field (string): The escape character for unenclosed fields. Defaults to "\\".
        trim_space (bool): Whether to trim spaces from fields. Defaults to False.
        field_optionally_enclosed_by (string): A character that may optionally enclose fields.
        null_if (list): A list of string values that should be interpreted as NULL.
        error_on_column_count_mismatch (bool, required): Whether to raise an error on column count mismatch. Defaults to True.
        replace_invalid_characters (bool): Whether to replace invalid characters. Defaults to False.
        empty_field_as_null (bool): Whether to treat empty fields as NULL.
        skip_byte_order_mark (bool): Whether to skip the byte order mark. Defaults to True.
        encoding (string): The file encoding. Defaults to "UTF8".
        comment (string): A comment regarding the file format.

    Python:

        ```python
        csv_file_format = CSVFileFormat(
            name="some_csv_file_format",
            compression=Compression.GZIP,
            field_delimiter=',',
            file_extension='csv',
            parse_header=True
        )
        ```

    Yaml:

        ```yaml
        csv_file_formats:
          - name: some_csv_file_format
            compression: GZIP
            field_delimiter: ','
            file_extension: 'csv'
            parse_header: true
        ```
    """

    resource_type = ResourceType.FILE_FORMAT
    props = Props(
        type=EnumProp("type", [FileType.CSV]),
        compression=EnumProp("compression", Compression),
        record_delimiter=StringProp("record_delimiter", alt_tokens=["NONE"]),
        field_delimiter=StringProp("field_delimiter", alt_tokens=["NONE"]),
        file_extension=StringProp("file_extension"),
        parse_header=BoolProp("parse_header"),
        skip_header=IntProp("skip_header"),
        skip_blank_lines=BoolProp("skip_blank_lines"),
        date_format=StringProp("date_format", alt_tokens=["AUTO"]),
        time_format=StringProp("time_format", alt_tokens=["AUTO"]),
        timestamp_format=StringProp("timestamp_format", alt_tokens=["AUTO"]),
        binary_format=EnumProp("binary_format", BinaryFormat),
        escape=StringProp("escape", alt_tokens=["NONE"]),
        escape_unenclosed_field=StringProp("escape_unenclosed_field", alt_tokens=["NONE"]),
        trim_space=BoolProp("trim_space"),
        field_optionally_enclosed_by=StringProp("field_optionally_enclosed_by", alt_tokens=["NONE"]),
        null_if=StringListProp("null_if", parens=True),
        error_on_column_count_mismatch=BoolProp("error_on_column_count_mismatch"),
        replace_invalid_characters=BoolProp("replace_invalid_characters"),
        empty_field_as_null=BoolProp("empty_field_as_null"),
        skip_byte_order_mark=BoolProp("skip_byte_order_mark"),
        encoding=StringProp("encoding", alt_tokens=["UTF8"]),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _CSVFileFormat

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        compression: Compression = None,
        record_delimiter: str = "\n",
        field_delimiter: str = None,
        file_extension: str = None,
        parse_header: bool = False,
        skip_header: int = None,
        skip_blank_lines: bool = False,
        date_format: str = "AUTO",
        time_format: str = "AUTO",
        timestamp_format: str = "AUTO",
        binary_format: BinaryFormat = BinaryFormat.HEX,
        escape: str = None,
        escape_unenclosed_field: str = "\\",
        trim_space: bool = False,
        field_optionally_enclosed_by: str = None,
        null_if: list[str] = None,
        error_on_column_count_mismatch: bool = True,
        replace_invalid_characters: bool = False,
        empty_field_as_null: bool = None,
        skip_byte_order_mark: bool = True,
        encoding: str = "UTF8",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _CSVFileFormat = _CSVFileFormat(
            name=self._name,
            owner=owner,
            compression=compression,
            record_delimiter=record_delimiter,
            field_delimiter=field_delimiter,
            file_extension=file_extension,
            parse_header=parse_header,
            skip_header=skip_header,
            skip_blank_lines=skip_blank_lines,
            date_format=date_format,
            time_format=time_format,
            timestamp_format=timestamp_format,
            binary_format=binary_format,
            escape=escape,
            escape_unenclosed_field=escape_unenclosed_field,
            trim_space=trim_space,
            field_optionally_enclosed_by=field_optionally_enclosed_by,
            null_if=null_if,
            error_on_column_count_mismatch=error_on_column_count_mismatch,
            replace_invalid_characters=replace_invalid_characters,
            empty_field_as_null=empty_field_as_null,
            skip_byte_order_mark=skip_byte_order_mark,
            encoding=encoding,
            comment=comment,
        )


# @_fix_class_documentation
# class CSVFileFormat(SchemaScoped, Resource):
#     """
#     CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY | VOLATILE } ] FILE FORMAT [ IF NOT EXISTS ] <name>
#       TYPE = CSV
#       [ formatTypeOptions ]
#       [ COMMENT = '<string_literal>' ]

#     formatTypeOptions ::=
#     -- If TYPE = CSV
#          COMPRESSION = AUTO | GZIP | BZ2 | BROTLI | ZSTD | DEFLATE | RAW_DEFLATE | NONE
#          RECORD_DELIMITER = '<character>' | NONE
#          FIELD_DELIMITER = '<character>' | NONE
#          FILE_EXTENSION = '<string>'
#          PARSE_HEADER = TRUE | FALSE
#          SKIP_HEADER = <integer>
#          SKIP_BLANK_LINES = TRUE | FALSE
#          DATE_FORMAT = '<string>' | AUTO
#          TIME_FORMAT = '<string>' | AUTO
#          TIMESTAMP_FORMAT = '<string>' | AUTO
#          BINARY_FORMAT = HEX | BASE64 | UTF8
#          ESCAPE = '<character>' | NONE
#          ESCAPE_UNENCLOSED_FIELD = '<character>' | NONE
#          TRIM_SPACE = TRUE | FALSE
#          FIELD_OPTIONALLY_ENCLOSED_BY = '<character>' | NONE
#          NULL_IF = ( '<string>' [ , '<string>' ... ] )
#          ERROR_ON_COLUMN_COUNT_MISMATCH = TRUE | FALSE
#          REPLACE_INVALID_CHARACTERS = TRUE | FALSE
#          EMPTY_FIELD_AS_NULL = TRUE | FALSE
#          SKIP_BYTE_ORDER_MARK = TRUE | FALSE
#          ENCODING = '<string>' | UTF8
#     """

#     resource_type = "FILE FORMAT"
# props = Props(
#     type=EnumProp("type", [FileType.CSV]),
#     compression=EnumProp("compression", Compression),
#     record_delimiter=StringProp("record_delimiter", alt_tokens=["NONE"]),
#     field_delimiter=StringProp("field_delimiter", alt_tokens=["NONE"]),
#     file_extension=StringProp("file_extension"),
#     parse_header=BoolProp("parse_header"),
#     skip_header=IntProp("skip_header"),
#     skip_blank_lines=BoolProp("skip_blank_lines"),
#     date_format=StringProp("date_format", alt_tokens=["AUTO"]),
#     time_format=StringProp("time_format", alt_tokens=["AUTO"]),
#     timestamp_format=StringProp("timestamp_format", alt_tokens=["AUTO"]),
#     binary_format=EnumProp("binary_format", BinaryFormat),
#     escape=StringProp("escape", alt_tokens=["NONE"]),
#     escape_unenclosed_field=StringProp("escape_unenclosed_field", alt_tokens=["NONE"]),
#     trim_space=BoolProp("trim_space"),
#     field_optionally_enclosed_by=StringProp("field_optionally_enclosed_by", alt_tokens=["NONE"]),
#     null_if=StringListProp("null_if", parens=True),
#     error_on_column_count_mismatch=BoolProp("error_on_column_count_mismatch"),
#     replace_invalid_characters=BoolProp("replace_invalid_characters"),
#     empty_field_as_null=BoolProp("empty_field_as_null"),
#     skip_byte_order_mark=BoolProp("skip_byte_order_mark"),
#     encoding=StringProp("encoding", alt_tokens=["UTF8"]),
#     comment=StringProp("comment"),
# )

#     name: str
#     owner: str = "SYSADMIN"
#     type: FileType = FileType.CSV
#     compression: Compression = None
#     record_delimiter: str = None
#     field_delimiter: str = None
#     file_extension: str = None
#     parse_header: bool = None
#     skip_header: int = None
#     skip_blank_lines: bool = None
#     date_format: str = None
#     time_format: str = None
#     timestamp_format: str = None
#     binary_format: BinaryFormat = None
#     escape: str = None
#     escape_unenclosed_field: str = None
#     trim_space: bool = None
#     field_optionally_enclosed_by: str = None
#     null_if: List[str] = []
#     error_on_column_count_mismatch: bool = None
#     replace_invalid_characters: bool = None
#     empty_field_as_null: bool = None
#     skip_byte_order_mark: bool = None
#     encoding: str = None
#     comment: str = None


# @_fix_class_documentation
# class JSONFileFormat(SchemaScoped, Resource):
#     """
#     CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY | VOLATILE } ] FILE FORMAT [ IF NOT EXISTS ] <name>
#       TYPE = JSON
#       [ formatTypeOptions ]
#       [ COMMENT = '<string_literal>' ]

#     formatTypeOptions ::=
#     -- If TYPE = JSON
#          COMPRESSION = AUTO | GZIP | BZ2 | BROTLI | ZSTD | DEFLATE | RAW_DEFLATE | NONE
#          DATE_FORMAT = '<string>' | AUTO
#          TIME_FORMAT = '<string>' | AUTO
#          TIMESTAMP_FORMAT = '<string>' | AUTO
#          BINARY_FORMAT = HEX | BASE64 | UTF8
#          TRIM_SPACE = TRUE | FALSE
#          NULL_IF = ( '<string>' [ , '<string>' ... ] )
#          FILE_EXTENSION = '<string>'
#          ENABLE_OCTAL = TRUE | FALSE
#          ALLOW_DUPLICATE = TRUE | FALSE
#          STRIP_OUTER_ARRAY = TRUE | FALSE
#          STRIP_NULL_VALUES = TRUE | FALSE
#          REPLACE_INVALID_CHARACTERS = TRUE | FALSE
#          IGNORE_UTF8_ERRORS = TRUE | FALSE
#          SKIP_BYTE_ORDER_MARK = TRUE | FALSE
#     """

#     resource_type = "FILE FORMAT"
#     props = Props(
#         type=EnumProp("type", [FileType.JSON]),
#         compression=EnumProp("compression", Compression),
#         date_format=StringProp("date_format", alt_tokens=["AUTO"]),
#         time_format=StringProp("time_format", alt_tokens=["AUTO"]),
#         timestamp_format=StringProp("timestamp_format", alt_tokens=["AUTO"]),
#         binary_format=EnumProp("binary_format", BinaryFormat),
#         trim_space=BoolProp("trim_space"),
#         null_if=StringListProp("null_if", parens=True),
#         file_extension=StringProp("file_extension"),
#         enable_octal=BoolProp("enable_octal"),
#         allow_duplicate=BoolProp("allow_duplicate"),
#         strip_outer_array=BoolProp("strip_outer_array"),
#         strip_null_values=BoolProp("strip_null_values"),
#         replace_invalid_characters=BoolProp("replace_invalid_characters"),
#         ignore_utf8_errors=BoolProp("ignore_utf8_errors"),
#         skip_byte_order_mark=BoolProp("skip_byte_order_mark"),
#         comment=StringProp("comment"),
#     )

#     name: str
#     owner: str = "SYSADMIN"
#     type: FileType = FileType.JSON
#     compression: Compression = None
#     date_format: str = None
#     time_format: str = None
#     timestamp_format: str = None
#     binary_format: BinaryFormat = None
#     trim_space: bool = None
#     null_if: List[str] = []
#     file_extension: str = None
#     enable_octal: bool = None
#     allow_duplicate: bool = None
#     strip_outer_array: bool = None
#     strip_null_values: bool = None
#     replace_invalid_characters: bool = None
#     ignore_utf8_errors: bool = None
#     skip_byte_order_mark: bool = None
#     comment: str = None


# @_fix_class_documentation
# class AvroFileFormat(SchemaScoped, Resource):
#     """
#     CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY | VOLATILE } ] FILE FORMAT [ IF NOT EXISTS ] <name>
#       TYPE = AVRO
#       [ formatTypeOptions ]
#       [ COMMENT = '<string_literal>' ]

#     formatTypeOptions ::=
#     -- If TYPE = AVRO
#          COMPRESSION = AUTO | GZIP | BROTLI | ZSTD | DEFLATE | RAW_DEFLATE | NONE
#          TRIM_SPACE = TRUE | FALSE
#          REPLACE_INVALID_CHARACTERS = TRUE | FALSE
#          NULL_IF = ( '<string>' [ , '<string>' ... ] )
#     """

#     resource_type = "FILE FORMAT"
#     props = Props(
#         type=EnumProp("type", [FileType.AVRO]),
#         compression=EnumProp("compression", Compression),
#         trim_space=BoolProp("trim_space"),
#         replace_invalid_characters=BoolProp("replace_invalid_characters"),
#         null_if=StringListProp("null_if", parens=True),
#         comment=StringProp("comment"),
#     )

#     name: str
#     owner: str = "SYSADMIN"
#     type: FileType = FileType.AVRO
#     compression: Compression = None
#     trim_space: bool = None
#     replace_invalid_characters: bool = None
#     null_if: List[str] = []
#     comment: str = None


# @_fix_class_documentation
# class OrcFileFormat(SchemaScoped, Resource):
#     """
#     CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY | VOLATILE } ] FILE FORMAT [ IF NOT EXISTS ] <name>
#       TYPE = ORC
#       [ formatTypeOptions ]
#       [ COMMENT = '<string_literal>' ]

#     formatTypeOptions ::=
#     -- If TYPE = ORC
#          TRIM_SPACE = TRUE | FALSE
#          REPLACE_INVALID_CHARACTERS = TRUE | FALSE
#          NULL_IF = ( '<string>' [ , '<string>' ... ] )
#     """

#     resource_type = "FILE FORMAT"
#     props = Props(
#         type=EnumProp("type", [FileType.ORC]),
#         trim_space=BoolProp("trim_space"),
#         replace_invalid_characters=BoolProp("replace_invalid_characters"),
#         null_if=StringListProp("null_if", parens=True),
#         comment=StringProp("comment"),
#     )

#     name: str
#     owner: str = "SYSADMIN"
#     type: FileType = FileType.ORC
#     trim_space: bool = None
#     replace_invalid_characters: bool = None
#     null_if: List[str] = []
#     comment: str = None


# @_fix_class_documentation
# class ParquetFileFormat(SchemaScoped, Resource):
#     """
#     CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY | VOLATILE } ] FILE FORMAT [ IF NOT EXISTS ] <name>
#       TYPE = PARQUET
#       [ formatTypeOptions ]
#       [ COMMENT = '<string_literal>' ]

#     formatTypeOptions ::=
#     -- If TYPE = PARQUET
#          COMPRESSION = AUTO | LZO | SNAPPY | NONE
#          SNAPPY_COMPRESSION = TRUE | FALSE
#          BINARY_AS_TEXT = TRUE | FALSE
#          TRIM_SPACE = TRUE | FALSE
#          REPLACE_INVALID_CHARACTERS = TRUE | FALSE
#          NULL_IF = ( '<string>' [ , '<string>' ... ] )
#     """

#     resource_type = "FILE FORMAT"
#     props = Props(
#         type=EnumProp("type", [FileType.PARQUET]),
#         compression=EnumProp("compression", [Compression.AUTO, Compression.LZO, Compression.SNAPPY, Compression.NONE]),
#         snappy_compression=BoolProp("snappy_compression"),
#         binary_as_text=BoolProp("binary_as_text"),
#         trim_space=BoolProp("trim_space"),
#         replace_invalid_characters=BoolProp("replace_invalid_characters"),
#         null_if=StringListProp("null_if", parens=True),
#         comment=StringProp("comment"),
#     )

#     name: str
#     owner: str = "SYSADMIN"
#     type: FileType = FileType.PARQUET
#     compression: Compression = None
#     snappy_compression: bool = None
#     binary_as_text: bool = None
#     trim_space: bool = None
#     replace_invalid_characters: bool = None
#     null_if: List[str] = []
#     comment: str = None


# @_fix_class_documentation
# class XMLFileFormat(SchemaScoped, Resource):
#     """
#     CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY | VOLATILE } ] FILE FORMAT [ IF NOT EXISTS ] <name>
#       TYPE = XML
#       [ formatTypeOptions ]
#       [ COMMENT = '<string_literal>' ]

#     formatTypeOptions ::=
#     -- If TYPE = XML
#          COMPRESSION = AUTO | GZIP | BZ2 | BROTLI | ZSTD | DEFLATE | RAW_DEFLATE | NONE
#          IGNORE_UTF8_ERRORS = TRUE | FALSE
#          PRESERVE_SPACE = TRUE | FALSE
#          STRIP_OUTER_ELEMENT = TRUE | FALSE
#          DISABLE_SNOWFLAKE_DATA = TRUE | FALSE
#          DISABLE_AUTO_CONVERT = TRUE | FALSE
#          REPLACE_INVALID_CHARACTERS = TRUE | FALSE
#          SKIP_BYTE_ORDER_MARK = TRUE | FALSE
#     """

#     resource_type = "FILE FORMAT"
#     props = Props(
#         type=EnumProp("type", [FileType.XML]),
#         compression=EnumProp("compression", Compression),
#         ignore_utf8_errors=BoolProp("ignore_utf8_errors"),
#         preserve_space=BoolProp("preserve_space"),
#         strip_outer_element=BoolProp("strip_outer_element"),
#         disable_snowflake_data=BoolProp("disable_snowflake_data"),
#         disable_auto_convert=BoolProp("disable_auto_convert"),
#         replace_invalid_characters=BoolProp("replace_invalid_characters"),
#         skip_byte_order_mark=BoolProp("skip_byte_order_mark"),
#         comment=StringProp("comment"),
#     )

#     name: str
#     owner: str = "SYSADMIN"
#     type: FileType = FileType.XML
#     compression: Compression = None
#     ignore_utf8_errors: bool = None
#     preserve_space: bool = None
#     strip_outer_element: bool = None
#     disable_snowflake_data: bool = None
#     disable_auto_convert: bool = None
#     replace_invalid_characters: bool = None
#     skip_byte_order_mark: bool = None
#     comment: str = None


# FileTypeMap = {
#     FileType.CSV: CSVFileFormat,
#     FileType.JSON: JSONFileFormat,
#     FileType.AVRO: AvroFileFormat,
#     FileType.ORC: OrcFileFormat,
#     FileType.PARQUET: ParquetFileFormat,
#     FileType.XML: XMLFileFormat,
# }


# class FileFormatProp(Prop):
#     """
#     FILE_FORMAT = my_named_ff
#     FILE_FORMAT = (FORMAT_NAME = my_named_ff)
#     FILE_FORMAT = (TYPE = CSV ...)
#     """

#     def __init__(self, label):
#         value_expr = (
#             _in_parens(IdentifierProp("format_name").parser("prop_value"))
#             | pp.original_text_for(pp.nested_expr())("prop_value")
#             | FullyQualifiedIdentifier("prop_value")
#             | StringLiteral("prop_value")
#         )

#         super().__init__(label, value_expr, eq=True)

#     def typecheck(self, prop_value):
#         # Prop is an identifier name
#         if isinstance(prop_value, list):
#             return ".".join(prop_value)

#         prop_value = prop_value.strip("()")
#         file_type = EnumProp("type", FileType).parse(prop_value)

#         if file_type is None:
#             raise ValueError(f"Invalid inline file format: {prop_value}")

#         # Prop is an inline file format
#         file_type_cls = FileTypeMap[file_type]
#         return _parse_props(file_type_cls.props, prop_value)

#     def render(self, value):
#         if value is None:
#             return ""
#         eq = " = " if self.eq else " "
#         if isinstance(value, dict):
#             value_str = ", ".join(f"{k} = {v}" for k, v in value.items())
#             value_str = f"({value_str})"
#         else:
#             value_str = str(value)
#         return f"{self.label}{eq}{value_str}"


# class FileFormat(Resource, ABC):
#     def __new__(
#         cls, type: Union[str, FileType], **kwargs
#     ) -> Union[CSVFileFormat, JSONFileFormat, AvroFileFormat, OrcFileFormat, ParquetFileFormat, XMLFileFormat]:
#         file_type = FileType.parse(type)
#         file_type_cls = FileTypeMap[file_type]
#         return file_type_cls(type=file_type, **kwargs)

#     @classmethod
#     def from_sql(cls, sql):
#         resource_cls = Resource.classes[_resolve_resource_class(sql)]
#         return resource_cls.from_sql(sql)

FileFormatMap = {
    FileType.CSV: CSVFileFormat,
    # FileType.JSON: JSONFileFormat,
    # FileType.AVRO: AvroFileFormat,
    # FileType.ORC: OrcFileFormat,
    # FileType.PARQUET: ParquetFileFormat,
    # FileType.XML: XMLFileFormat,
}


def _resolver(data: dict):
    return FileFormatMap[FileType(data["type"])]


Resource.__resolvers__[ResourceType.FILE_FORMAT] = _resolver



from dataclasses import dataclass, field
from typing import Any, Union

from inflection import singularize

from .resource import Resource, ResourcePointer, ResourceSpec
from .role import Role
from .user import User
from ..enums import ParseableEnum, ResourceType
from ..identifiers import FQN, resource_label_for_type, resource_type_for_label
from ..parse import _parse_grant
from ..privs import GlobalPriv, GLOBAL_PRIV_DEFAULT_OWNERS, _all_privs_for_resource_type
from ..props import Props, FlagProp, IdentifierProp
from ..resource_name import ResourceName
from ..scope import AccountScope

# TODO: Should Grant objects verify grant types in advance?


@dataclass(unsafe_hash=True)
class _Grant(ResourceSpec):
    priv: str
    on: ResourceName
    on_type: ResourceType
    to: Role
    grant_option: bool = False
    owner: Role = None
    _privs: list[str] = field(default_factory=list, metadata={"forces_add": True})

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.priv, str):
            self.priv = self.priv.upper()
        if not self._privs:
            if self.priv == "ALL":
                self._privs = sorted(_all_privs_for_resource_type(self.on_type))
            else:
                self._privs = [self.priv]


class Grant(Resource):
    """
    Description:
        Represents a grant of privileges on a resource to a role in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/grant-privilege

    Fields:
        priv (string, required): The privilege to grant. Examples include 'SELECT', 'INSERT', 'CREATE TABLE'.
        on (string or Resource, required): The resource on which the privilege is granted. Can be a string like 'ACCOUNT' or a specific resource object.
        to (string or Role, required): The role to which the privileges are granted.
        grant_option (bool): Specifies whether the grantee can grant the privileges to other roles. Defaults to False.
        owner (string or Role): The owner role of the grant. Defaults to 'SYSADMIN'.

    Python:

        ```python
        # Global Privs:
        grant = Grant(priv="CREATE WAREHOUSE", on="ACCOUNT", to="somerole")

        # Warehouse Privs:
        grant = Grant(priv="OPERATE", on=Warehouse(name="foo"), to="somerole")
        grant = Grant(priv="OPERATE", on_warehouse="foo", to="somerole")

        # Schema Privs:
        grant = Grant(priv="CREATE TABLE", on=Schema(name="foo"), to="somerole")
        grant = Grant(priv="CREATE TABLE", on_schema="foo", to="somerole")

        # Table Privs:
        grant = Grant(priv="SELECT", on_table="sometable", to="somerole")
        ```

    Yaml:

        ```yaml
        - Grant:
            priv: "SELECT"
            on: "some_table"
            to: "some_role"
            grant_option: true
        ```
    """

    resource_type = ResourceType.GRANT
    props = Props(
        to=IdentifierProp("to", eq=False, consume="role"),
        grant_option=FlagProp("with grant option"),
    )
    scope = AccountScope()
    spec = _Grant

    def __init__(
        self,
        priv: str = None,
        on: Any = None,
        to: Role = None,
        grant_option: bool = False,
        owner: str = None,
        **kwargs,
    ):

        kwargs.pop("_privs", None)

        priv = priv.value if isinstance(priv, ParseableEnum) else priv

        # Handle instantiation from data dict
        on_type = kwargs.pop("on_type", None)
        if on_type:
            on_type = ResourceType(on_type)

        # Collect on_ kwargs
        on_kwargs = {}
        for keyword, arg in kwargs.copy().items():
            if keyword.startswith("on_"):
                on_kwargs[keyword] = kwargs.pop(keyword)

        # Handle dynamic on_ kwargs
        if on_kwargs:
            for keyword, arg in on_kwargs.items():
                if on is not None:
                    raise ValueError("You can only specify one 'on' parameter, multiple found")
                elif keyword.startswith("on_all"):
                    raise ValueError("You must use GrantOnAll for all grants")
                elif keyword.startswith("on_future"):
                    raise ValueError("You must use FutureGrant for future grants")
                else:
                    # Grant targeting a specific resource
                    # on_{resource} kwargs
                    # on_schema="foo" -> on=Schema(name="foo")
                    on = arg
                    on_type = resource_type_for_label(keyword[3:])
        # Handle on= kwarg
        else:
            if on is None:
                raise ValueError("You must specify an 'on' parameter")
            elif isinstance(on, ResourcePointer):
                on_type = on.resource_type
                on = on.name
            elif isinstance(on, Resource):
                on_type = on.resource_type
                on = on._data.name
            elif isinstance(on, str) and on.upper() == "ACCOUNT":
                on = "ACCOUNT"
                on_type = ResourceType.ACCOUNT

        if owner is None:
            if on_type == ResourceType.ACCOUNT and isinstance(priv, GlobalPriv):
                owner = GLOBAL_PRIV_DEFAULT_OWNERS.get(priv, "SYSADMIN")
            # Hacky fix
            elif on_type == ResourceType.SCHEMA and on.upper().startswith("SNOWFLAKE"):
                owner = "ACCOUNTADMIN"
            elif "INTEGRATION" in str(on_type):
                owner = "ACCOUNTADMIN"
            else:
                owner = "SYSADMIN"

        if to is None and kwargs.get("to_role"):
            to = kwargs.pop("to_role")

        super().__init__(**kwargs)
        self._data: _Grant = _Grant(
            priv=priv,
            on=on,
            on_type=on_type,
            to=to,
            grant_option=grant_option,
            owner=owner,
        )

        self.requires(self._data.to)
        granted_on = None
        if on_type:
            granted_on = ResourcePointer(name=on, resource_type=on_type)
            self.requires(granted_on)

    def __repr__(self):  # pragma: no cover
        priv = getattr(self._data, "priv", "")
        on = getattr(self._data, "on", "")
        to = getattr(self._data, "to", "")
        return f"{self.__class__.__name__}(priv={priv}, on={on}, to={to})"

    @classmethod
    def from_sql(cls, sql):
        parsed = _parse_grant(sql)
        return cls(**parsed)

    @property
    def fqn(self):
        return grant_fqn(self._data)

    @property
    def on(self) -> str:
        return self._data.on

    @property
    def on_type(self) -> ResourceType:
        return self._data.on_type

    @property
    def to(self):
        return self._data.to

    @property
    def priv(self):
        return self._data.priv


def grant_fqn(grant: _Grant):
    on = f"{resource_label_for_type(grant.on_type)}/{grant.on}"
    # if grant.on_type == ResourceType.ACCOUNT:
    #     on = "ACCOUNT"
    return FQN(
        name=grant.to.name,
        params={
            "priv": grant.priv,
            "on": on,
        },
    )


@dataclass(unsafe_hash=True)
class _FutureGrant(ResourceSpec):
    priv: str
    on_type: ResourceType
    in_type: ResourceType
    in_name: ResourceName
    to: Role
    grant_option: bool = False

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.priv, str):
            self.priv = self.priv.upper()


class FutureGrant(Resource):
    """
    GRANT <privilege> ON FUTURE <on_type> IN <in_type> <in_name> TO <to> [ WITH GRANT OPTION ]
    """

    resource_type = ResourceType.FUTURE_GRANT
    props = Props(
        priv=IdentifierProp("priv", eq=False),
        on_type=IdentifierProp("on type", eq=False),
        database=IdentifierProp("database", eq=False),
        to=IdentifierProp("to", eq=False),
    )
    scope = AccountScope()
    spec = _FutureGrant

    def __init__(
        self,
        priv: str,
        to: Role,
        grant_option: bool = False,
        owner: str = None,
        **kwargs,
    ):
        """
        Usage
        -----

        Database Object Privs:
        >>> Grant(priv="CREATE TABLE", on_future_schemas_in=Database(name="somedb"), to="somerole")
        >>> Grant(priv="CREATE TABLE", on_future_schemas_in_database="somedb", to="somerole")

        Schema Object Privs:
        >>> Grant(priv="SELECT", on_future_tables_in=Schema(name="someschema"), to="somerole")
        >>> Grant(priv="READ", on_future_image_repositories_in_schema="someschema", to="somerole")

        """
        on_type = kwargs.pop("on_type", None)
        in_type = kwargs.pop("in_type", None)
        in_name = kwargs.pop("in_name", None)
        granted_in_ref = None

        if all([on_type, in_type, in_name]):
            in_type = ResourceType(in_type)
            on_type = ResourceType(on_type)
            granted_in_ref = ResourcePointer(name=in_name, resource_type=in_type)

        else:

            # Collect on_ kwargs
            on_kwargs = {}
            for keyword, _ in kwargs.copy().items():
                if keyword.startswith("on_future_"):
                    on_kwargs[keyword] = kwargs.pop(keyword)

            if len(on_kwargs) != 1:
                raise ValueError("You must specify one 'on_future_' parameter")

            # Handle on_future_ kwargs
            if on_kwargs:
                for keyword, arg in on_kwargs.items():

                    # At some point we need to support _in_sometype=SomeType(blah)

                    if isinstance(arg, Resource):
                        on_type = ResourceType(singularize(keyword[10:-3]))
                        in_type = arg.resource_type
                        in_name = str(arg.fqn)
                        granted_in_ref = arg
                    else:
                        on_stmt, in_stmt = keyword.split("_in_")
                        on_type = ResourceType(singularize(on_stmt[10:]))
                        in_type = ResourceType(in_stmt)
                        in_name = arg
                        granted_in_ref = ResourcePointer(name=in_name, resource_type=in_type)

        super().__init__(**kwargs)
        self._data: _FutureGrant = _FutureGrant(
            priv=priv,
            on_type=on_type,
            in_type=in_type,
            in_name=in_name,
            to=to,
            grant_option=grant_option,
        )
        self.requires(self._data.to)
        if granted_in_ref:
            self.requires(granted_in_ref)

    @classmethod
    def from_sql(cls, sql):
        parsed = _parse_grant(sql)
        return cls(**parsed)

    @property
    def fqn(self):
        return future_grant_fqn(self._data)

    @property
    def priv(self) -> str:
        return self._data.priv

    @property
    def on_type(self) -> ResourceType:
        return self._data.on_type

    @property
    def in_type(self) -> ResourceType:
        return self._data.in_type

    @property
    def in_name(self) -> str:
        return self._data.in_name

    @property
    def to(self):
        return self._data.to


def future_grant_fqn(data: _FutureGrant):
    in_type = resource_label_for_type(data.in_type)
    in_name = data.in_name
    on_type = resource_label_for_type(data.on_type).upper()
    return FQN(
        name=data.to.name,
        params={
            "priv": data.priv,
            "on": f"{in_type}/{in_name}.<{on_type}>",
        },
    )


@dataclass(unsafe_hash=True)
class _GrantOnAll(ResourceSpec):
    priv: str
    on_type: ResourceType
    in_type: ResourceType
    in_name: str
    to: Role
    grant_option: bool = False
    owner: Role = "SECURITYADMIN"

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.priv, str):
            self.priv = self.priv.upper()
        if self.in_type not in [ResourceType.DATABASE, ResourceType.SCHEMA]:
            raise ValueError("in_type must be either DATABASE or SCHEMA")


class GrantOnAll(Resource):
    """
    GRANT
          { schemaPrivileges         | ALL [ PRIVILEGES ] } ON ALL SCHEMAS IN DATABASE <db_name>
        | { schemaObjectPrivileges   | ALL [ PRIVILEGES ] } ON ALL <object_type_plural> IN { DATABASE <db_name> | SCHEMA <schema_name> }
        }
    TO [ ROLE ] <role_name> [ WITH GRANT OPTION ]
    """

    resource_type = ResourceType.GRANT_ON_ALL
    props = Props(
        priv=IdentifierProp("priv", eq=False),
        on_type=IdentifierProp("on type", eq=False),
        in_type=IdentifierProp("in type", eq=False),
        in_name=IdentifierProp("in name", eq=False),
        to=IdentifierProp("to", eq=False),
        grant_option=FlagProp("with grant option"),
    )
    scope = AccountScope()
    spec = _GrantOnAll

    def __init__(
        self,
        priv: str,
        to: Role,
        grant_option: bool = False,
        **kwargs,
    ):
        """
        Usage
        -----

        Schema Privs:
        >>> GrantOnAll(priv="CREATE TABLE", on_all_schemas_in_database="somedb", to="somerole")
        >>> GrantOnAll(priv="CREATE VIEW", on_all_schemas_in=Database(name="somedb"), to="somerole")

        Schema Object Privs:
        >>> GrantOnAll(priv="SELECT", on_all_tables_in_schema="sch", to="somerole")
        >>> GrantOnAll(priv="SELECT", on_all_views_in_database="somedb", to="somerole")

        """
        on_type = kwargs.pop("on_type", None)
        in_type = kwargs.pop("in_type", None)
        in_name = kwargs.pop("in_name", None)

        # Init from serialized
        if all([on_type, in_type, in_name]):
            in_type = ResourceType(in_type)
            on_type = ResourceType(on_type)
        else:

            # Collect on_ kwargs
            on_kwargs = {}
            for keyword, _ in kwargs.copy().items():
                if keyword.startswith("on_all_"):
                    on_kwargs[keyword] = kwargs.pop(keyword)

            if len(on_kwargs) != 1:
                raise ValueError("You must specify one 'on_all_' parameter")

            # Handle on_all_ kwargs
            if on_kwargs:
                for keyword, arg in on_kwargs.items():
                    on_keyword = keyword.split("_")[2]
                    on_type = ResourceType(singularize(on_keyword))
                    if isinstance(arg, Resource):
                        in_type = arg.resource_type
                        in_name = str(arg.fqn)
                    else:
                        in_stmt = keyword.split("_in_")[1]
                        in_type = ResourceType(in_stmt)
                        in_name = arg

        if in_type == ResourceType.SCHEMA and in_name.upper().startswith("SNOWFLAKE"):
            owner = "ACCOUNTADMIN"
        else:
            owner = "SECURITYADMIN"

        super().__init__(**kwargs)
        self._data: _GrantOnAll = _GrantOnAll(
            priv=priv,
            on_type=on_type,
            in_type=in_type,
            in_name=in_name,
            to=to,
            grant_option=grant_option,
            owner=owner,
        )
        self.requires(self._data.to)

    @classmethod
    def from_sql(cls, sql):
        parsed = _parse_grant(sql)
        return cls(**parsed)

    @property
    def fqn(self):
        return grant_on_all_fqn(self._data)


def grant_on_all_fqn(grant: _GrantOnAll):
    return FQN(
        name=grant.to.name,
        params={
            "on_type": str(grant.on_type),
            "in_type": str(grant.in_type),
            "in_name": grant.in_name,
        },
    )


@dataclass(unsafe_hash=True)
class _RoleGrant(ResourceSpec):
    role: Role
    to_role: Role = None
    to_user: User = None
    # owner: str = "SECURITYADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.to_role is not None and self.to_user is not None:
            raise ValueError("You can only grant to a role or a user, not both")
        if self.to_role is None and self.to_user is None:
            raise ValueError("You must specify a role or a user to grant to")


class RoleGrant(Resource):
    """
    GRANT ROLE <name> TO { ROLE <parent_role_name> | USER <user_name> }
    """

    resource_type = ResourceType.ROLE_GRANT
    props = Props(
        role=IdentifierProp("role", eq=False),
        to_role=IdentifierProp("to role", eq=False),
        to_user=IdentifierProp("to user", eq=False),
    )
    scope = AccountScope()
    spec = _RoleGrant

    def __init__(
        self,
        role: Role,
        to_role: Role = None,
        to_user: User = None,
        # owner: str = None,  # = "SECURITYADMIN"
        **kwargs,
    ):
        """
        Usage
        -----

        Grant to Role:
        >>> RoleGrant(role="somerole", to_role="someotherrole")
        >>> RoleGrant(role="somerole", to=Role(name="someuser"))

        Grant to User:
        >>> RoleGrant(role="somerole", to_user="someuser")
        >>> RoleGrant(role="somerole", to=User(name="someuser"))
        """

        to = kwargs.pop("to", None)
        if to:
            if to_role or to_user:
                raise ValueError("You can only grant to a role or a user, not both")
            if isinstance(to, Role):
                to_role = to
            elif isinstance(to, User):
                to_user = to
            else:
                raise ValueError("You can only grant to a role or a user")

        super().__init__(**kwargs)
        self._data: _RoleGrant = _RoleGrant(
            role=role,
            to_role=to_role,
            to_user=to_user,
            # owner=owner,
        )
        self.requires(
            self._data.role,
            self._data.to_role,
            self._data.to_user,
        )

    @classmethod
    def from_sql(cls, sql):
        props = _parse_grant(sql)
        return RoleGrant(**props)

    def __repr__(self):  # pragma: no cover
        role = getattr(self._data, "role", "")
        to_role = getattr(self._data, "to_role", "")
        to_user = getattr(self._data, "to_user", "")
        to = to_role or to_user
        return f"RoleGrant(role={role}, to={to})"

    # @property
    # def fqn(self):
    #     subject = "user" if self._data.to_user else "role"
    #     name = self._data.to_user.name if self._data.to_user else self._data.to_role.name
    #     return FQN(
    #         name=self._data.role.name,
    #         params={subject: name},
    #     )

    @property
    def fqn(self):
        return role_grant_fqn(self._data)

    @property
    def role(self) -> Role:
        return self._data.role

    @property
    def to(self) -> Union[Role, User]:
        return self._data.to_role or self._data.to_user


def role_grant_fqn(role_grant: _RoleGrant):
    subject = "user" if role_grant.to_user else "role"
    name = role_grant.to_user.name if role_grant.to_user else role_grant.to_role.name
    return FQN(
        name=role_grant.role.name,
        params={subject: name},
    )



import pytest

from titan.blueprint import Action, Blueprint, ResourceChange
from titan.parse import parse_URN


@pytest.fixture(scope="session")
def remote_state(removed_db):
    state = {
        parse_URN("urn::XYZ123:database/EXISTING_DB"): {
            "name": "EXISTING_DB",
            "transient": False,
            "owner": "SYSADMIN",
        },
    }
    state.update(removed_db)
    return state


@pytest.fixture(scope="session")
def new_db():
    return {
        parse_URN("urn::XYZ123:database/NEW_DB"): {
            "name": "NEW_DB",
            "transient": False,
            "owner": "SYSADMIN",
        }
    }


@pytest.fixture(scope="session")
def changed_db():
    return {
        parse_URN("urn::XYZ123:database/EXISTING_DB"): {
            "name": "EXISTING_DB",
            "transient": False,
            "owner": "SYSADMIN",
            "default_ddl_collation": "UTF8",
        }
    }


@pytest.fixture(scope="session")
def removed_db():
    return {
        parse_URN("urn::XYZ123:database/REMOVED_DB"): {
            "name": "REMOVED_DB",
            "transient": False,
            "owner": "SYSADMIN",
        }
    }


@pytest.fixture(scope="session")
def manifest(new_db, changed_db):
    manifest = {
        "_urns": [
            parse_URN("urn::XYZ123:database/EXISTING_DB"),
            parse_URN("urn::XYZ123:database/NEW_DB"),
        ],
        "_refs": [],
    }
    manifest.update(new_db)
    manifest.update(changed_db)
    return manifest


def test_plan_add_action(remote_state, manifest, new_db):
    bp = Blueprint()
    changes = bp._plan(remote_state, manifest)
    urn, after = new_db.popitem()
    expected = ResourceChange(Action.ADD, urn, {}, after, after)
    assert expected in changes


def test_plan_change_action(remote_state, manifest, changed_db):
    bp = Blueprint()
    changes = bp._plan(remote_state, manifest)
    urn, data = changed_db.popitem()
    delta = {"default_ddl_collation": "UTF8"}
    expected = ResourceChange(Action.CHANGE, urn, remote_state[urn], data, delta)
    assert expected in changes


def test_plan_remove_action(remote_state, manifest, removed_db):
    bp = Blueprint()
    changes = bp._plan(remote_state, manifest)
    urn, _ = removed_db.popitem()
    expected = ResourceChange(Action.REMOVE, urn, remote_state[urn], {}, {})
    assert expected in changes
